#! /usr/bin/env python

# "DShield.py"
# dshield.org linux iptables log parser by Eelco Lempsink
#
# $Id: dshield.py,v 1.0 2001/11/29 20:50:37 eelcol Exp $
#
# Parses and mails an iptables log (generated by the 2.4.x Linux
# kernel) for submission to DShield.org to the DShield format
#
# Copyright (c) 2001 Eelco Lempsink (eelcolempsink@gmx.net)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# See http://www.gnu.org/licenses/gpl.txt or COPYING


import ConfigParser, fileinput, getopt, os, re, smtplib, string, sys, time

idstr = '$Id: dshield.py,v 1.0 2001/11/29 20:50:37 eelcol Exp $'
client = 'DShield.py'
version = string.split(idstr)[2]

months = {'Jan':'01', 'Feb':'02', 'Mar':'03', 'Apr':'04', 'May':'05', 'Jun':'06', 'Jul':'07', 'Aug':'08', 'Sep':'09', 'Oct':'10', 'Nov':'11', 'Dec':'12'}

flags = ['SYN', 'ACK', 'URG', 'RST', 'PSH', 'FIN']


def main():
    "The 'heart' of the program, executes all other functions"

    global options
    options = read_options()

    subject = "FORMAT DSHIELD USERID %s TZ %s %s %s #" % (options['userid'], options['timez'], client, version)
    report, count, SRC, SPT, DST, DPT, PROTO, FLAGS = ('',) * 8

    lines = 0
    time.clock()
    
    if options['verbose'] != 'no': print "Opening %s for reading..." % options['file']
    for line in fileinput.input(options['file']):
        if string.count(line, 'IN=') and not string.count(line, 'ACCEPT'):
            month, day, ltime, count, SRC, SPT, DST, DPT, PROTO, FLAGS = parse_iptables(line, count, SRC, SPT, DST, DPT, PROTO, FLAGS)
            report = "%s%s-%s-%s %s %s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\r\n" % (report, options['year'], month, day, ltime, options['timez'], options['userid'], count, SRC, SPT, DST, DPT, PROTO, FLAGS)
            lines = lines + 1

    
    if lines:
        clocked = time.clock()
        if options['verbose'] != 'no': print "Parsed %d lines in %.1f second(s) (%.1f lines/s)" % (lines, clocked, lines/clocked)
        mail(subject, report)
        if options['verbose'] != 'no': print "Mailed the report to %s" % options['mailto']
    else:
        if options['verbose'] != 'no': print "No lines found, bummer..."


def read_options():
    """Read options from config files and the command line, returns the 
    defaults if no user options are found"""

    year = time.localtime(time.time())[0]
    if time.localtime(time.time())[8]:
        timez = -time.altzone / 3600
    else:
        timez = -time.timezone / 3600
    if timez >= 0:
        timez = '+' + string.zfill(timez, 2) + ':00'
    else:
        timez = string.zfill(timez, 3) + ':00'

    # defaults
    options = {'file'   : '/var/log/syslog.0',
               'year'   : year,
               'smtp'   : 'localhost',
               'userid' : '0',
               'timez'  : timez,
               'email'  : 'postmaster@localhost',
               'mailto' : 'postmaster@localhost',
               'copy'   : '',
               'obfusc' : '',
               'verbose': ''}

    # read from config files
    config = ConfigParser.ConfigParser()

    configfile = 'dshieldpy.conf'
    config.read(['/etc/' + configfile, os.environ['HOME'] + '/' + configfile, configfile])

    for option in options.keys():
        execcode = "if config.defaults().has_key('%s'): options['%s'] = config.defaults()['%s']" % (option, option, option)
        exec execcode

    # read from command line
    helpstr = 'Usage: ' + sys.argv[0] + ' [OPTIONS]' + """\n
  * -f, --file <filename>   the firewall log that must be parsed
    -y, --year <year>       the current year (since it's not logged) as YYYY
    -s, --smtp <host>       the SMTP server to use
  * -u, --userid <id>       your DShield.org userid (number)
    -t, --timez <timezone>  your timezone in the format +HH:MM or -HH:MM
  * -e, --email <email>     your email address (as registered at DShield.org)
  * -m, --mailto <email>    use reports@dshield.org for actual submissions
    -c, --copy              send the report also to yourself (set by --email)
    -o, --obfusc <p|c>      set to 'partial' or 'complete' for IP obfuscation
    -h, --help              show this help
    -v, --verbose           show some stats
    -V, --version           show version info

A '*' means this option (probably) has to be set for correct submission."""

    optlist, args = getopt.getopt(sys.argv[1:], 'f:y:s:u:t:e:m:co:hVv', ['file=', 'year=', 'smtp=', 'userid=', 'timez=', 'email=', 'mailto=', 'copy', 'obfusc=', 'help', 'version', 'verbose'])

    for o, a in optlist:
        if (o == '-c' or o == '--copy'): 
            options['copy'] = 'yes'
        elif (o == '-h' or o == '--help'):
            print helpstr
            sys.exit()
        elif (o == '-V' or o == '--version'):
            print client, version
            sys.exit()
        elif (o == '-v' or o == '--verbose'):
            options['verbose'] = 'yes'
        else:
            for option in options.keys():
                execcode = "if (o == '-%s' or o == '--%s'): options['%s'] = a" % (option[0], option, option)
                exec execcode

    #return all options
    return options


def mail(subject, report):
    "Mail the report to the 'mailto' address as set in the options"
    
    if options['copy'] != 'no':
        mailto = (options['mailto'], options['email'])
    else:
        mailto = (options['mailto'],)

    msg = "From: %s\r\nTo: %s\r\nSubject: %s\r\n\r\n%s" % (options['email'], string.join(mailto, ', '), subject, report)

    mailserv = smtplib.SMTP(options['smtp'])
    mailserv.sendmail(options['email'], mailto, msg)
    mailserv.quit()


def parse_iptables(line, pcount, pSRC, pSPT, pDST, pDPT, pPROTO, pFLAGS):
    "Parse an iptables log line"

    splitted = re.split(r'\s+', line)

    month = months[splitted[0]]
    day   = string.zfill(splitted[1],2)
    ltime = splitted[2]

    PROTO = re.search(r'PROTO=\w+',line).group()[6:]

    if PROTO == '':
        PROTO = '???'

    # icmp
    if PROTO == 'ICMP':
        SPT = re.search(r'TYPE=\d+', line).group()[5:]
        DPT = re.search(r'CODE=\d+', line).group()[5:]
    else:
        SPT = re.search(r'SPT=\d+',line).group()[4:]
        DPT = re.search(r'DPT=\d+',line).group()[4:]

    SRC = re.search(r'SRC=\d+\.\d+\.\d+\.\d+',line).group()[4:]
    DST = re.search(r'DST=\d+\.\d+\.\d+\.\d+',line).group()[4:]

    # tcp flags
    FLAGS = ''
    if PROTO == 'TCP':
        for flag in flags:
            if splitted.count(flag) > 0:
                FLAGS = FLAGS + flag[0]

    # counting
    if SRC == pSRC and SPT == pSPT and DST == pDST and DPT == pDPT and PROTO == pPROTO and FLAGS == pFLAGS:
        count = pcount + 1
    else:
        count = 1

    # obfuscation
    obfusc = options['obfusc']
    if (obfusc == 'partial' or obfusc == 'p'):
        DST = re.sub(r'\d+', '10', DST, 1)
    elif (obfusc == 'complete' or obfusc == 'c'):
        DST = '10.0.0.1'

    return month, day, ltime, count, SRC, SPT, DST, DPT, PROTO, FLAGS


main()
